#include "Camera.h"

const float Camera::incX = 0.5f;
const float Camera::incY = 0.005f;
const float Camera::initJumpSpeed = 0.3f;
const float Camera::gravity = -.0005f;
const float Camera::groundHeight = 30.0f;
const float Camera::moveSpeed = 0.1f;

Camera::Camera(){
	eye			= vec4d(0.0, groundHeight, 100.0, 1.0);
	lookAt		= vec4d(0.0, 0.0, -1.0, 1.0);
	up			= vec4d(0.0, 1.0, 0.0, 1.0);
	speed		= vec4d(0.0, 0.0, 0.0, 1.0);
	moveState	= 0x0000;
	xzAngle		= 90.0f;
	isJumping	= false;
}

Camera::~Camera(){};

/*
 * Rotate the camera according to mouse movement. Since the up vector will
 * always be (0,1,0), lookAt.y is clamped so that the lookAt vector
 * will not coincide with the up vector.
 */
void Camera::rotate(float dX, float dY){
	xzAngle = fmodf(xzAngle + incX*dX, 360.0f);
	lookAt.y += incY * dY;
	if(lookAt.y >= 1.0f)	lookAt.y = .999f;
	if(lookAt.y <= -1.0f)	lookAt.y = -.999f;
}

void Camera::jump(){
	if(!isJumping){
		speed.y = initJumpSpeed;
		isJumping = true;
	}
}

/*
 * If the camera is jumping, the xz speed is frozen
 * and the y speed is affected by gravity. Else,
 * x and z speed is calculated and added to eye position.
 * Also, the eye is kept at a min. distance above ground.
 */
void Camera::update(int msDelta){
	float x = cos(PI * xzAngle/180.0f);
	float z = -sin(PI * xzAngle/180.0f);
	lookAt.x = x * (1 - fabs(lookAt.y));
	lookAt.z = z * (1 - fabs(lookAt.y));

	if(isJumping)
		speed.y += gravity * msDelta;
	else{
		speed = vec4d(0.0, 0.0, 0.0, 0.0);
		x *= moveSpeed;
		z *= moveSpeed;

		if(moveState & CAMERA_FORWARD){
			speed.x += x;
			speed.z += z;
		}
		if(moveState & CAMERA_BACK){
			speed.x -= x;
			speed.z -= z;
		}
		if(moveState & CAMERA_LEFT){
			speed.x += z;
			speed.z -= x;
		}
		if(moveState & CAMERA_RIGHT){
			speed.x -= z;
			speed.z += x;
		}
	}
	eye += speed * msDelta;
	if(eye.y < groundHeight){
		eye.y = groundHeight;
		speed.y = 0;
		isJumping = false;
	}
}

/*
 * Since the lookAt vector is treated as a direction and not
 * as a point, it's temporarily added to the eye to correct it.
 */
void Camera::draw(){
	lookAt += eye;
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eye.x, eye.y, eye.z, 
			  lookAt.x, lookAt.y, lookAt.z, 
			  up.x, up.y, up.z);
	lookAt -= eye;
}