/*
 * Mat4 is stored in row-major order.
 */

#ifndef Mat4_H
#define Mat4_H

#include "Vec4.h"

template<class T> class Mat4{
private:
	T mi[4][4];

public:
	inline Mat4<T>(const Mat4<T> &r);
	inline Mat4<T>(T m00=0, T m01=0, T m02=0, T m03=0,
			T m10=0, T m11=0, T m12=0, T m13=0,
			T m20=0, T m21=0, T m22=0, T m23=0,
			T m30=0, T m31=0, T m32=0, T m33=0);
	inline ~Mat4<T>();
	inline T *operator[](int index);
	inline operator T*();
	inline bool operator==(const Mat4<T> &r);
	inline bool operator!=(const Mat4<T> &r);
	inline Mat4<T> &operator=(const Mat4<T> &r);
	template<class U> inline Mat4<T> &operator=(const Mat4<U> &r);
	template<class U> inline Mat4<T> &operator=(const U r[16]);
	template<class U> inline Mat4<T> &operator+=(const Mat4<U> &r);
	template<class U> inline Mat4<T> operator+(const Mat4<U> &r)const;
	template<class U> inline Mat4<T> &operator-=(const Mat4<U> &r);
	template<class U> inline Mat4<T> operator-(const Mat4<U> &r)const;
	template<class U> inline Mat4<T> &operator*=(const Mat4<U> &r);
	template<class U> inline Mat4<T> operator*(const Mat4<U> &r)const;
	template<class U> inline Vec4<U> operator*(const Vec4<U> &r)const;
	inline Mat4<T> &operator*=(double scalar);
	inline Mat4<T> operator*(double scalar)const;
	inline Mat4<T> &operator/=(double scalar);
	inline Mat4<T> operator/(double scalar)const;
	inline Mat4<T> &transpose();
	std::string toString();
};

template<class T>
inline Mat4<T>::Mat4(const Mat4<T> &r){
	*this = r;
}

template<class T>
inline Mat4<T>::Mat4(T m00=0, T m01=0, T m02=0, T m03=0,
			  T m10=0, T m11=0, T m12=0, T m13=0,
			  T m20=0, T m21=0, T m22=0, T m23=0,
			  T m30=0, T m31=0, T m32=0, T m33=0){
	mi[0][0] = m00; mi[0][1] = m01; mi[0][2] = m02; mi[0][3] = m03;
	mi[1][0] = m10; mi[1][1] = m11; mi[1][2] = m12; mi[1][3] = m13;
	mi[2][0] = m20; mi[2][1] = m21; mi[2][2] = m22; mi[2][3] = m23;
	mi[3][0] = m30; mi[3][1] = m31; mi[3][2] = m32; mi[3][3] = m33;
}

template<class T>
inline Mat4<T>::~Mat4(){
}	

template<class T>
inline T* Mat4<T>::operator[](int index){
	return mi[index];
}

template<class T>
inline Mat4<T>::operator T*(){
	return &mi[0][0];
}

template<class T>
inline bool Mat4<T>::operator==(const Mat4 &r){
	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			if(mi[i][j] != r.mi[i][j])
				return false;
	return true;
}

template<class T>
inline bool Mat4<T>::operator!=(const Mat4 &r){
	return !((*this)==r);
}

template<class T>
inline Mat4<T> &Mat4<T>::operator=(const Mat4<T> &r){
	if(this != &r)
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
				mi[i][j] = r.mi[i][j];
	return *this;
}

template<class T> template<class U>
inline Mat4<T> &Mat4<T>::operator=(const Mat4<U> &r){
	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			mi[i][j] = (T)r.mi[i][j];
	return *this;
}

template<class T> template<class U>
inline Mat4<T> &Mat4<T>::operator=(const U r[16]){
	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			mi[i][j] = (T)r[i*4 + j];
	return *this;
}

template<class T> template<class U>
inline Mat4<T> &Mat4<T>::operator+=(const Mat4<U> &r){
	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			mi[i][j] += (T)r.mi[i][j];
	return *this;
}

template<class T> template<class U>
inline Mat4<T> Mat4<T>::operator+(const Mat4<U> &r)const{
	return Mat4(*this) += r;
}

template<class T> template<class U>
inline Mat4<T> &Mat4<T>::operator-=(const Mat4<U> &r){
	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			mi[i][j] -= (T)r.mi[i][j];
	return *this;
}

template<class T> template<class U>
inline Mat4<T> Mat4<T>::operator-(const Mat4<U> &r)const{
	return Mat4(*this) -= r;
}

template<class T> template<class U>
inline Mat4<T> &Mat4<T>::operator*=(const Mat4<U> &r){
	Mat4<T> temp;
	for(int i = 0; i < 4; i++){
		for(int j = 0; j < 4; j++){
			T tot = 0;
			for(int k = 0; k < 4; k++)
				tot += mi[i][k]*(T)r.mi[k][j];
			temp[i][j] = tot;
		}
	}
	*this = temp;
	return *this;
}

template<class T> template<class U>
inline Mat4<T> Mat4<T>::operator*(const Mat4<U> &r)const{
	return Mat4(*this) *= r;
}

//Vec4 is treated as 4x1 matrix on rhs
template<class T> template <class U>
inline Vec4<U> Mat4<T>::operator*(const Vec4<U> &r)const{
	Vec4<U> final;
	for(int i = 0; i < 4; i++){
		T tot = 0;
		for(int j = 0; j < 4; j++)
			tot += mi[i][j] * (T)r.v[j];
		final[i] = tot;
	}
	return final;
}

template<class T>
inline Mat4<T> &Mat4<T>::operator*=(double scalar){
	for(int i = 0; i < 4; i++)
		for(int j = 0; j < 4; j++)
			mi[i][j] = (T) (mi[i][j] * scalar);
	return *this;
}

template<class T>
inline Mat4<T> Mat4<T>::operator*(double scalar)const{
	return Mat4(*this) *= scalar;
}

template<class T>
inline Mat4<T> &Mat4<T>::operator/=(double scalar){
	if(scalar != 0){
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
				mi[i][j] = (T) (mi[i][j] / scalar);
	}
	return *this;
}

template<class T>
inline Mat4<T> Mat4<T>::operator/(double scalar)const{
	return Mat4(*this) /= scalar;
}

template<class T>
inline Mat4<T> &Mat4<T>::transpose(){
	T temp;
	for(int i = 0; i < 3; i++)
		for(int j = 1; j < 4; j++)
			if(j > i){
				temp = mi[i][j];
				mi[i][j] = mi[j][i];
				mi[j][i] = temp;
			}
	return *this;
}

template<class T>
std::string Mat4<T>::toString(){
	std::ostringstream oss;
	for(int i = 0; i < 4; i++){
		for(int j = 0; j < 4; j++){
			oss << mi[i][j];
			if(j != 3) 
				oss << "\t";
		}
		if(i != 3)
			oss << "\n";
	}
	return oss.str();
}

template<class T>
inline Mat4<T> operator*(double scalar, Mat4<T> &r){
	return r * scalar;
}

template<class T>
std::ostream &operator<<(std::ostream &stream, Mat4<T> &r){
	return stream << r.toString();
}

typedef Mat4<float>		Mat4f;
typedef Mat4<double>	Mat4d;
typedef Mat4<int>		Mat4i;

#endif //Mat4_H