#include "scene.h"
using namespace std;

Scene::Scene(){
	texture = new GLuint[NUM_TEXTURES];
	cubeObj = new CubeObj();
	skybox = new Skybox();
	ground = new Ground();
	camera = new Camera();
	cubeObj->position		= vec4d(0.0, 30.0, 0.0, 1.0);
	ground->position		= vec4d(0.0, 0.0, 0.0, 1.0);
	
	setup();
}

Scene::~Scene(){
	glDeleteTextures(NUM_TEXTURES, texture);
	delete texture;
	delete cubeObj;
	delete skybox;
	delete ground;
	delete camera;
	texture = NULL;
	cubeObj = NULL;
	skybox = NULL;
	ground = NULL;
	camera = NULL;
}

void Scene::draw(){
	camera->draw();
	skybox->draw();
	drawReflection();
	ground->draw();
	cubeObj->draw();
}

/*
 * Updates the camera position, and clamps it to the boundaries of the ground.
 */
void Scene::update(int msDelta){
	camera->update(msDelta);
	camera->eye.x = (camera->eye.x < -ground->getSize()) ? -ground->getSize() :
		(camera->eye.x > ground->getSize() ? ground->getSize() : camera->eye.x);
	camera->eye.z = (camera->eye.z < -ground->getSize()) ? -ground->getSize() :
		(camera->eye.z > ground->getSize() ? ground->getSize() : camera->eye.z);
	skybox->position = camera->eye;
	cubeObj->update(msDelta);
}

void Scene::setup(){
	GLuint displayList = glGenLists(4);

	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	loadTextures();
	cubeObj->setup(displayList, texture);
	skybox->setup(displayList, texture);
	ground->setup(displayList, texture);
	initLights();
	initMaterials();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void Scene::drawReflection(){
		mainLight.position.y *= -1;
		mainLight.set();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glScalef(1.0f, -1.0f, 1.0f);

		glEnable(GL_STENCIL_TEST);
		glStencilFunc(GL_NEVER, 0, 0);			
		glStencilOp(GL_INCR, GL_INCR, GL_INCR);
		ground->draw();
		glStencilFunc(GL_EQUAL, 1, 1);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		glFrontFace(GL_CW);
		glPushMatrix();
			glTranslatef(0.0f, -2.0f * skybox->position.y, 0.0f);
			skybox->draw();
		glPopMatrix();
		cubeObj->draw();

		glDisable(GL_STENCIL_TEST);
		glFrontFace(GL_CCW);
		glPopMatrix();
		mainLight.position.y *= -1;
		mainLight.set();
}

void Scene::loadTextures(){
	char *data, *filenames[NUM_TEXTURES];
	FILE *file;
	filenames[CUBE_TEXTURE]		= "Images\\wood.rgb";
	filenames[SPHERE_TEXTURE]	= "Images\\reflective.rgb";
	filenames[GROUND_TEXTURE]	= "Images\\stone.rgb";
	filenames[POS_X]			= "Images\\pos_x.rgb";
	filenames[POS_Y]			= "Images\\pos_y.rgb";
	filenames[POS_Z]			= "Images\\pos_z.rgb";
	filenames[NEG_X]			= "Images\\neg_x.rgb";
	filenames[NEG_Y]			= "Images\\neg_y.rgb";
	filenames[NEG_Z]			= "Images\\neg_z.rgb";
	
	data = (char *)malloc(512*512*3);
	glGenTextures(NUM_TEXTURES, texture);

	for(int i = 0; i < NUM_TEXTURES; i++){
		file = fopen(filenames[i], "rb");
		if(file == NULL)
			printf("%s could not be read", filenames[i]);
		else{
			fread(data, 512*512*3, 1, file);
			fclose(file);
			glBindTexture(GL_TEXTURE_2D, texture[i]);
			if(i == GROUND_TEXTURE){
				glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
				glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			}
			else{
				glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
				glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
			}
			gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 512, 512, GL_RGB, GL_UNSIGNED_BYTE, data);
		}
	}
	free(data);
}

void Scene::initLights(){
	glEnable(GL_LIGHTING);
	mainLight.lightNum	= GL_LIGHT0;
	mainLight.position	= vec4f(-80.0f, 120.0f, -50.0f, 1.0f);
	mainLight.ambient	= vec4f(0.3f, 0.3f, 0.3f, 1.0f);
	mainLight.diffuse	= vec4f(0.7f, 0.7f, 0.7f, 1.0f);
	mainLight.specular	= vec4f(1.0f, 1.0f, 1.0f, 1.0f);
	mainLight.set();
	mainLight.enable();
}

void Scene::initMaterials(){
	glEnable(GL_NORMALIZE);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
	glMateriali(GL_FRONT, GL_SHININESS, 64);
}