/*
 * To avoid normalization problems, addition, subtraction,
 * scalar multiplication & division, dot & cross products
 * are calculated using xyz only, not w.
 */

#ifndef Vec4_H
#define Vec4_H

#include <iostream>
#include <sstream>
#include <math.h>
#include <string>
#include <assert.h>

template <class T> class Vec4{

public:
	union{
		struct {T x, y, z, w;};
		struct {T r, g, b, a;};
		struct {T s, t, p, q;};
		T v[4];
	};

	//constructors
	inline Vec4<T>(T x=0, T y=0, T z=0, T w=0);
	inline Vec4<T>(const Vec4<T> &r);
	inline ~Vec4<T>();

	inline T &operator[](int index);
	inline operator T*();
	inline bool operator==(const Vec4<T> &r)const;
	inline bool operator!=(const Vec4<T> &r)const;
	inline Vec4<T> &operator=(const Vec4<T> &r);
	template<class U> inline Vec4<T> &operator=(const Vec4<U> &r);
	template<class U> inline Vec4<T> &operator=(const U r[4]);
	template<class U> inline Vec4<T> &operator+=(const Vec4<U> &r);
	template<class U> inline Vec4<T> operator+(const Vec4<U> &r)const;
	template<class U> inline Vec4<T> &operator-=(const Vec4<U> &r);
	template<class U> inline Vec4<T> operator-(const Vec4<U> &r)const;
	template<class U> inline Vec4<T> &operator*=(const Vec4<U> &r);
	template<class U> inline Vec4<T> operator*(const Vec4<U> &r)const;
	template<class U> inline Vec4<T> &operator/=(const Vec4<U> &r);
	template<class U> inline Vec4<T> operator/(const Vec4<U> &r)const;
	inline Vec4<T> &operator*=(double scalar);
	inline Vec4<T> operator*(double scalar)const;
	inline Vec4<T> &operator/=(double scalar);
	inline Vec4<T> operator/(double scalar)const;

	inline T magnitude()const;
	inline Vec4<T> &normalize();
	inline Vec4<T> &divideW();
	std::string toString();
};

template<class T>
inline Vec4<T>::Vec4(T x=0, T y=0, T z=0, T w=0) :
	x(x), y(y), z(z), w(w){
}

template<class T>
inline Vec4<T>::Vec4(const Vec4<T> &r) :
	x(r.x), y(r.y), z(r.z), w(r.w){
}

template<class T>
inline Vec4<T>::~Vec4(){
}

template<class T>
inline T &Vec4<T>::operator[](int index){
	assert(index <= 3);
	return v[index];
}

template<class T>
inline Vec4<T>::operator T*(){
	return &x;
}

template<class T>
inline bool Vec4<T>::operator==(const Vec4<T> &r)const{
	if(x == r.x && y == r.y && z == r.z && w == r.w)
		return true;
	return false;
}

template<class T>
inline bool Vec4<T>::operator!=(const Vec4<T> &r)const{
	return !((*this)==r);
}

template<class T>
inline Vec4<T> &Vec4<T>::operator=(const Vec4<T> &r){
	if(this != &r){
		x = r.x;
		y = r.y;
		z = r.z;
		w = r.w;
	}
	return *this;
}

template<class T> template<class U>
inline Vec4<T> &Vec4<T>::operator=(const Vec4<U> &r){
	x = (T)r.x;
	y = (T)r.y;
	z = (T)r.z;
	w = (T)r.w;
	return *this;
}

template<class T> template<class U>
inline Vec4<T> &Vec4<T>::operator=(const U r[4]){
	if(v != r)
		for(int i = 0; i < 4; i++)
			v[i] = (T)r[i];
	return *this;
}

//xyz only
template<class T> template<class U>
inline Vec4<T> &Vec4<T>::operator+=(const Vec4<U> &r){
	x += (T)r.x;
	y += (T)r.y;
	z += (T)r.z;
	return *this;
}

//xyz only
template<class T> template<class U>
inline Vec4<T> Vec4<T>::operator+(const Vec4<U> &r)const{
	return Vec4(*this) += r;
}

//xyz only
template<class T> template<class U>
inline Vec4<T> &Vec4<T>::operator-=(const Vec4<U> &r){
	x -= (T)r.x;
	y -= (T)r.y;
	z -= (T)r.z;
	return *this;
}	

//xyz only
template<class T> template<class U>
inline Vec4<T> Vec4<T>::operator-(const Vec4<U> &r)const{
	return Vec4(*this) -= r;
}

//xyz only
template<class T> template<class U> 
inline Vec4<T> &Vec4<T>::operator*=(const Vec4<U> &r){
	x *= (T)r.x;
	y *= (T)r.y;
	z *= (T)r.z;
	return *this;
}

//xyz only
template<class T> template<class U> 
inline Vec4<T> Vec4<T>::operator*(const Vec4<U> &r)const{
	return Vec4(*this) *= r;
}

//xyz only
template<class T> template<class U> 
inline Vec4<T> &Vec4<T>::operator/=(const Vec4<U> &r){
	x /= (T)r.x;
	y /= (T)r.y;
	z /= (T)r.z;
	return *this;
}

//xyz only
template<class T> template<class U> 
inline Vec4<T> Vec4<T>::operator/(const Vec4<U> &r)const{
	return Vec4(*this) /= r;
}

//xyz only
template<class T>
inline Vec4<T> &Vec4<T>::operator*=(double scalar){
	x = (T) (x * scalar);
	y = (T) (y * scalar);
	z = (T) (z * scalar);
	return *this;
}

//xyz only
template<class T>
inline Vec4<T> Vec4<T>::operator*(double scalar)const{
	return Vec4(*this) *= scalar;
}

//xyz only
template<class T>
inline Vec4<T> &Vec4<T>::operator/=(double scalar){
	assert(scalar != 0);
	x = (T) (x / scalar);
	y = (T) (y / scalar);
	z = (T) (z / scalar);
	return *this;
}

//xyz only
template<class T>
inline Vec4<T> Vec4<T>::operator/(double scalar)const{
	return Vec4(*this) /= scalar;
}

//xyz only--NEEDS TO BE FLOAT/DOUBLE TO BE ACCURATE
template<class T>
inline T Vec4<T>::magnitude()const{
	return (T)sqrt((double)(x*x + y*y + z*z));
}

//xyz only
template<class T>
inline Vec4<T> &Vec4<T>::normalize(){
	T d = magnitude();
	x /= d;
	y /= d;
	z /= d;
	return *this;
}

template<class T>
inline Vec4<T> &Vec4<T>::divideW(){
	*this /= w;
	w = 1;
	return *this;
}

template<class T>
std::string Vec4<T>::toString(){
	std::ostringstream oss;
	oss << "(";
	for(int i = 0; i < 4; i++){
		oss << v[i];
		if(i != 3)
			oss << ", ";
	}
	oss << ")";
	return oss.str();
}

//xyz only
template<class T, class U>
inline T dot(const Vec4<T> &l, const Vec4<U> &r){
	return l.x*r.x + l.y*r.y + l.z*r.z;
}

//xyz only
template<class T, class U>
inline Vec4<T> cross(const Vec4<T> &l, const Vec4<U> &r){
	Vec4<T> temp = Vec4<T>(0,0,0,1);
	temp.x = l.y*r.z + l.z*r.y;
	temp.y = l.z*r.x + l.x*r.z;
	temp.z = l.x*r.y + l.y*r.x;
	return temp;
}

template<class T>
inline Vec4<T> operator*(double scalar, Vec4<T> &r){
	return r * scalar;
}

template<class T>
inline std::ostream &operator <<(std::ostream &stream, Vec4<T> &r){
	return stream << r.toString();
}

typedef Vec4<float>		Vec4f;
typedef Vec4<double>	Vec4d;
typedef Vec4<int>		Vec4i;
typedef Vec4<float>		Color4f;
typedef Vec4<double>	Color4d;
typedef Vec4<int>		Color4i;

#endif // Vec4_H